/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <pkgmgr-info.h>
#include <pkgmgr_installer_info.h>
#include <aul.h>

#include "log.h"
#include "utils.h"
#include "pkgmgr_parser_plugin_interface.h"

#include <wait.h>
#include <dirent.h>
#include <sys/stat.h>

#include <algorithm>
#include <string>

#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <unistd.h>

#include "common.h"

#ifdef  LOG_TAG
#undef  LOG_TAG
#endif
#define LOG_TAG "NETCORE_INSTALLER_PLUGIN"

#ifndef DEVICE_API_DIR
#error "DEVICE_API_DIR is missed"
#endif

#ifndef RUNTIME_DIR
#error "RUNTIME_DIR is missed"
#endif

#ifndef CROSSGEN_PATH
#error "CROSSGEN_PATH is missed"
#endif

#define __XSTR(x) #x
#define __STR(x) __XSTR(x)
static const char* DeviceAPIDir = __STR(DEVICE_API_DIR);
static const char* RuntimeDir = __STR(RUNTIME_DIR);
static const char* CrossgenPath = __STR(CROSSGEN_PATH);
static const char* JITPath = __STR(RUNTIME_DIR)"/libclrjit.so";
#undef __STR
#undef __XSTR

static void crossgen(const char* dll_path, const char* app_path);
static void smack_(const char* dll_path, const char* label);

std::string Replace(std::string &str, const std::string& from, const std::string& to)
{
  size_t start_pos = 0;
  while((start_pos = str.find(from, start_pos)) != std::string::npos)
  {
    str.replace(start_pos, from.length(), to);
    start_pos += to.length();
  }
  return str;
}

void create_ni_platform()
{
  std::string corlib = ConcatPath(RuntimeDir, "System.Private.CoreLib.dll");
  std::string nicorlib = ConcatPath(RuntimeDir, "System.Private.CoreLib.ni.dll");

  if (FileNotExist(nicorlib))
  {
    crossgen(corlib.c_str(), nullptr);
    smack_(nicorlib.c_str(), "_");
  }

  const char* platform_dirs[] = {RuntimeDir, DeviceAPIDir, "/usr/bin"};
  const char* ignores[] = {corlib.c_str()};

  create_ni_under_dirs(platform_dirs, 3, ignores, 1, [](const char* ni){
      smack_(ni, "_");
  });
}

void create_ni_select(const char* dll_path)
{
  std::string corlib = ConcatPath(RuntimeDir, "System.Private.CoreLib.dll");
  std::string nicorlib = ConcatPath(RuntimeDir, "System.Private.CoreLib.ni.dll");

  if (FileNotExist(nicorlib))
  {
    crossgen(corlib.c_str(), nullptr);
    smack_(nicorlib.c_str(), "_");
  }

  if (!FileNotExist(dll_path))
  {
    std::string str_path = dll_path;
    std::string ni_path = Replace(str_path, std::string(".dll"), std::string(".ni.dll"));
    if (FileNotExist(ni_path))
      crossgen(dll_path, nullptr);
    else
      printf("Already [%s] file is exist\n", ni_path.c_str());
    smack_(ni_path.c_str(), "_");
  }
}

static void smack_(const char* dll_path, const char* label)
{
  static const char* CHKSMACK = "/usr/bin/chsmack";
  pid_t pid = fork();
  if (pid == -1)
  {
    return;
  }

  if (pid > 0)
  {
    int status;
    waitpid(pid, &status, 0);
    if (WIFEXITED(status))
    {
      return;
    }
  }
  else
  {
    const char* args[] = {
      CHKSMACK,
      "-a", label,
      dll_path,
      nullptr
    };
    execv(CHKSMACK, const_cast<char*const*>(args));

    exit(0);
  }
}

static void crossgen(const char* dll_path, const char* app_path)
{
  //pid_t parent = getpid();
  pid_t pid = fork();
  if (pid == -1)
  {
    return;
  }

  if (pid > 0)
  {
    int status;
    waitpid(pid, &status, 0);
    if (WIFEXITED(status))
    {
      return;
    }
  }
  else
  {
    // search dlls in the application directory first, to use application dlls
    // instead of system dlls when proceeding NI
    std::vector<std::string> tpaDir;
    if (app_path != NULL)
    {
      std::string path(app_path);
      std::string::size_type prev_pos = 0, pos = 0;
      while((pos = path.find(':', pos)) != std::string::npos)
      {
        std::string substring(path.substr(prev_pos, pos - prev_pos));
        tpaDir.push_back(substring);
        prev_pos = ++pos;
      }
      std::string substring(path.substr(prev_pos, pos - prev_pos));
      tpaDir.push_back(substring);
    }
    tpaDir.push_back(RuntimeDir);
    tpaDir.push_back(DeviceAPIDir);

    std::string tpa;
    AssembliesInDirectory(tpaDir, tpa);

    std::vector<const char*> argv =
    {
      CrossgenPath,
      "/Trusted_Platform_Assemblies", tpa.c_str(),
      "/JITPath", JITPath,
      "/FragileNonVersionable"
    };
    if (app_path != nullptr)
    {
      argv.push_back("/App_Paths");
      argv.push_back(app_path);
    }
    argv.push_back(dll_path);
    argv.push_back(nullptr);

    /*
    for (const char* arg : argv)
    {
      printf("%s ", arg);
    }
    printf("\n");
    */
    printf("+ %s\n", dll_path);

    execv(CrossgenPath, const_cast<char* const*>(argv.data()));
    exit(0);
  }
}

static int get_root_path(const char *pkgid, std::string& root_path)
{
  int ret = 0;
  char *path = 0;

  uid_t uid = 0;

  if (pkgmgr_installer_info_get_target_uid(&uid) < 0)
  {
    _ERR("Failed to get UID");
    return -1;
  }

  _INFO("user id is %d", uid);

  pkgmgrinfo_pkginfo_h handle;
  if (uid == 0)
  {
    ret = pkgmgrinfo_pkginfo_get_pkginfo(pkgid, &handle);
    if (ret != PMINFO_R_OK)
      return -1;
  }
  else
  {
    ret = pkgmgrinfo_pkginfo_get_usr_pkginfo(pkgid, uid, &handle);
    if (ret != PMINFO_R_OK)
      return -1;
  }

  ret = pkgmgrinfo_pkginfo_get_root_path(handle, &path);
  if (ret != PMINFO_R_OK) {
    pkgmgrinfo_pkginfo_destroy_pkginfo(handle);
    return -1;
  }
  root_path = path;
  pkgmgrinfo_pkginfo_destroy_pkginfo(handle);

  return 0;
}

static bool NIExist(const std::string& path, std::string& ni)
{
  static const char* possible_exts[] = {
    ".ni.dll", ".NI.dll", ".NI.DLL", ".ni.DLL",
    ".ni.exe", ".NI.exe", ".NI.EXE", ".ni.EXE"
  };
  std::string fname = path.substr(0, path.size() - 4);

  struct stat sb;

  for (const char* ext : possible_exts)
  {
    std::string f = fname + ext;
    if (stat(f.c_str(), &sb) == 0)
    {
      ni = f;
      return true;
    }
  }

  return false;
}

void create_ni_under_dirs(const char* root_paths[], int count, const char* ignores[], int igcount, after_create cb)
{
  std::string app_paths;
  for (int i=0; i<count; i++)
  {
    app_paths += root_paths[i];
    app_paths += ':';
  }
  if (app_paths.back() == ':')
    app_paths.pop_back();

  auto convert = [&app_paths, ignores, igcount, &cb](const char* path, const char* name)
  {
    for (int i=0; i<igcount; i++)
    {
      if (strcmp(path, ignores[i]) == 0)
        return;
    }
    std::string ni;
    if (IsManagedAssembly(path) && !IsNativeImage(path) && !NIExist(path, ni))
    {
      crossgen(path, app_paths.c_str());
      if (NIExist(path, ni))
      {
        // change owner and groups for generated ni file.
        struct stat info;
        if (!stat(path, &info))
        {
          if (chown(ni.c_str(), info.st_uid, info.st_gid) == -1)
          {
            _ERR("Failed to change owner and group name");
          }
        }

        if (cb != nullptr)
        {
          cb(ni.c_str());
        }
      }
    }
  };

  for (int i=0; i<count; i++)
  {
    ScanFilesInDir(root_paths[i], convert, -1);
  }
}
void create_ni_under_dirs(const char* root_paths[], int count, after_create cb)
{
  create_ni_under_dirs(root_paths, count, nullptr, 0, cb);
}
void create_ni_under_dirs(const char* root_paths[], int count)
{
  create_ni_under_dirs(root_paths, count, nullptr);
}

int create_ni_under_pkg_root(const char* pkg_name)
{
  std::string pkgroot;
  if (get_root_path(pkg_name, pkgroot) < 0)
  {
    return 1;
  }

  //printf("pkgroot : %s\n", pkgroot.c_str());

  std::string bindir = ConcatPath(pkgroot, "bin");
  std::string libdir = ConcatPath(pkgroot, "lib");
  
  //printf("bindir : %s\n", bindir.c_str());
  //printf("libdir : %s\n", libdir.c_str());
  _INFO("bindir : %s", bindir.c_str());
  _INFO("libdir : %s", libdir.c_str());

  const char* paths[] = {
    bindir.c_str(),
    libdir.c_str()
  };

  // change smack label for generated ni file.
  std::string label = "User::Pkg::" + std::string(pkg_name) + "::RO";
  create_ni_under_dirs(paths, 2, [label](const char* ni){
      smack_(ni, label.c_str());
  });

  return 0;
}
