/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

using System;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.Loader;
using System.Collections.Generic;

namespace Tizen.Runtime.Coreclr
{
    public class AssemblyLoader : AssemblyLoadContext
    {
        private const string NativeAssemblyInfix = ".ni";

        private const string DllAssemblySuffix = ".dll";

        private const string ExeAssemblySuffix = ".exe";

        private const string NativeDllAssemblySuffix = NativeAssemblyInfix + DllAssemblySuffix;

        private static readonly string[] s_suffixes = new string[] { NativeDllAssemblySuffix, DllAssemblySuffix, ExeAssemblySuffix };

        private SortedSet<string> _dllDirectories = new SortedSet<string>();
        private SortedSet<string> _nativeDirectories = new SortedSet<string>();

        private HashSet<FileInfo> _assemblyCache = new HashSet<FileInfo>();

        public AssemblyLoader()
        {
            AssemblyLoadContext.Default.Resolving += OnResolving;
        }

        public IEnumerable<string> DllDirectories
        {
            get { return _dllDirectories; }
        }

        public IEnumerable<string> NativeDirectories
        {
            get { return _nativeDirectories; }
        }

        public void AddSearchableDirectory(string directory)
        {
            if (Directory.Exists(directory))
            {
                _dllDirectories.Add(directory);
                _nativeDirectories.Add(directory);

                foreach (var file in Directory.GetFiles(directory))
                {
                    var info = new FileInfo(file);

                    if (s_suffixes.Contains(info.Extension))
                    {
                        _assemblyCache.Add(info);
                    }
                }
            }
        }

        public void RemoveSearchableDirectory(string directory)
        {
            _dllDirectories.Remove(directory);
            _nativeDirectories.Remove(directory);

            _assemblyCache.RemoveWhere(x => x.DirectoryName == directory);
        }

        public Assembly LoadFromPath(string path)
        {
            if (0 == string.Compare(path,  // strA
                                    path.Length - NativeDllAssemblySuffix.Length,  // indexA
                                    NativeAssemblyInfix,  // strB
                                    0,  // indexB
                                    NativeAssemblyInfix.Length,  // length
                                    StringComparison.OrdinalIgnoreCase))  // options
            {
                return LoadFromNativeImagePath(path, null);
            }
            else
            {
                return LoadFromAssemblyPath(path);
            }
        }

        protected override Assembly Load(AssemblyName assemblyName)
        {
            return Resolve(assemblyName);
        }

        protected override IntPtr LoadUnmanagedDll(string unmanagedDllName)
        {
            IntPtr native = base.LoadUnmanagedDll(unmanagedDllName);
            if (native == IntPtr.Zero)
            {
                foreach (string dir in NativeDirectories)
                {
                    FileInfo f = new FileInfo(Path.Combine(dir, unmanagedDllName));
                    if (File.Exists(f.FullName))
                    {
                        native = LoadUnmanagedDllFromPath(f.FullName);
                        break;
                    }
                }
            }

            return native;
        }

        private Assembly Resolve(AssemblyName assemblyName)
        {
            foreach (string suffix in s_suffixes)
            {
                var info = _assemblyCache.FirstOrDefault(x => x.Name == assemblyName.Name + suffix);

                if (info != null)
                {
                    return LoadFromPath(info.FullName);
                }
            }

            return null;
        }

        private Assembly OnResolving(AssemblyLoadContext context, AssemblyName assemblyName)
        {
            return Resolve(assemblyName);
        }
    }
}
